@@ -19,7 +19,7 @@ from ivy_vision.containers import CameraGeometry as _CameraGeometry
 MIN_DENOMINATOR = 1e-12
 
 
-def create_uniform_pixel_coords_image(image_dims, batch_shape=None, normalized=False, dev_str=None):
+def create_uniform_pixel_coords_image(image_dims, batch_shape=None, normalized=False, homogeneous=True, dev_str='cpu'):
     """
     Create image of homogeneous integer :math:`xy` pixel co-ordinates :math:`\mathbf{X}\in\mathbb{Z}^{h×w×3}`, stored
     as floating point values. The origin is at the top-left corner of the image, with :math:`+x` rightwards, and
@@ -34,8 +34,10 @@ def create_uniform_pixel_coords_image(image_dims, batch_shape=None, normalized=F
     :type image_dims: sequence of ints.
     :param batch_shape: Shape of batch. Assumed no batch dimensions if None.
     :type batch_shape: sequence of ints, optional
-    :param normalized: Whether to normalize x-y pixel co-ordinates to the range 0-1.
-    :type normalized: bool
+    :param normalized: Whether to normalize x-y pixel co-ordinates to the range 0-1. Default is False.
+    :type normalized: bool, optional
+    :param homogeneous: Whether the pixel co-ordinates should be 3D homogeneous or just 2D. Default is True.
+    :type: homogeneous: bool, optional
     :param dev_str: device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
     :type dev_str: str, optional
     :return: Image of homogeneous pixel co-ordinates *[batch_shape,height,width,3]*
@@ -48,7 +50,7 @@ def create_uniform_pixel_coords_image(image_dims, batch_shape=None, normalized=F
 
     # other shape specs
     num_batch_dims = len(batch_shape)
-    flat_shape = [1] * num_batch_dims + image_dims + [3]
+    flat_shape = [1] * num_batch_dims + image_dims + [2]
     tile_shape = batch_shape + [1] * 3
 
     # H x W x 1
@@ -66,12 +68,15 @@ def create_uniform_pixel_coords_image(image_dims, batch_shape=None, normalized=F
     if normalized:
         pixel_y_coords = pixel_y_coords / (float(image_dims[0]) + MIN_DENOMINATOR)
 
-    # H x W x 1
-    ones = _ivy.ones_like(pixel_x_coords, dev_str=dev_str)
+    # BS x H x W x 2
+    pix_coords = _ivy.tile(_ivy.reshape(_ivy.concatenate((pixel_x_coords, pixel_y_coords), -1), flat_shape), tile_shape)
 
-    # BS x H x W x 3
-    return _ivy.tile(_ivy.reshape(_ivy.concatenate((pixel_x_coords, pixel_y_coords, ones), -1),
-                                  flat_shape), tile_shape)
+    if homogeneous:
+        # BS x H x W x 3
+        pix_coords = _ivy_mec.make_coordinates_homogeneous(pix_coords)
+
+    # BS x H x W x 2or3
+    return pix_coords
 
 
 def persp_angles_to_focal_lengths(persp_angles, image_dims, dev_str=None):
@@ -241,45 +246,75 @@ def depth_to_ds_pixel_coords(depth, uniform_pixel_coords=None, batch_shape=None,
     return uniform_pixel_coords * depth
 
 
-def depth_to_radial_depth(depth, inv_calib_mat, uniform_pixel_coords=None, batch_shape=None, image_dims=None):
+def depth_to_radial_depth(depth, inv_calib_mat, uniform_pix_coords=None, batch_shape=None, image_dims=None):
     """
-    Get radial depth image :math:`\mathbf{X}_{rd}\in\mathbb{R}^{h×w×1}` from depth image
-    :math:`\mathbf{X}_d\in\mathbb{R}^{h×w×1}`.\n
+    Get radial depth image :math:`\mathbf{X}_{rd}\in\mathbb{R}^{hxw×1}` from depth image
+    :math:`\mathbf{X}_d\in\mathbb{R}^{hxw×1}`.\n
 
-    :param depth: Depth image *[batch_shape,h,w,1]*
+    :param depth: Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_dims,1]*
     :type depth: array
     :param inv_calib_mat: Inverse calibration matrix *[batch_shape,3,3]*
     :type inv_calib_mat: array
-    :param uniform_pixel_coords: Image of homogeneous pixel co-ordinates. Created if None. *[batch_shape,h,w,3]*
-    :type uniform_pixel_coords: array, optional
+    :param uniform_pix_coords: Uniform homogeneous pixel co-ordinates, constructed if None. *[batch_shape,image_dims,3]*
+    :type uniform_pix_coords: array, optional
     :param batch_shape: Shape of batch. Inferred from inputs if None.
     :type batch_shape: sequence of ints, optional
-    :param image_dims: Image dimensions. Inferred from inputs in None.
+    :param image_dims: Image shape. Inferred from inputs in None.
     :type image_dims: sequence of ints, optional
-    :return: Radial depth image *[batch_shape,h,w,1]*
+    :return: Radial depth image *[batch_shape,image_dims,1]*
     """
 
     if batch_shape is None:
-        batch_shape = depth.shape[:-3]
+        batch_shape = inv_calib_mat.shape[:-2]
+    num_batch_dims = len(batch_shape)
 
     if image_dims is None:
-        image_dims = depth.shape[-3:-1]
+        image_dims = depth.shape[num_batch_dims:-1]
 
     # shapes as list
     batch_shape = list(batch_shape)
     image_dims = list(image_dims)
 
     # BS x H x W x 3
-    if uniform_pixel_coords is None:
-        uniform_pixel_coords = create_uniform_pixel_coords_image(image_dims, batch_shape, dev_str=_ivy.dev_str(depth))
-
-    # BS x H x W x 3
-    ds_pixel_coords = depth_to_ds_pixel_coords(depth, uniform_pixel_coords, batch_shape, image_dims)
-
-    # BS x H x W x 3
-    cam_coords = ds_pixel_to_cam_coords(ds_pixel_coords, inv_calib_mat, batch_shape, image_dims)[..., 0:3]
+    if uniform_pix_coords is None:
+        uniform_pix_coords = create_uniform_pixel_coords_image(image_dims, batch_shape, dev_str=_ivy.dev_str(depth))
+    ds_pixel_coords = uniform_pix_coords * depth
 
     # BS x H x W x 1
+    return ds_pixel_coords_to_radial_depth(ds_pixel_coords, inv_calib_mat, batch_shape, image_dims)
+
+
+def ds_pixel_coords_to_radial_depth(ds_pixel_coords, inv_calib_mat, batch_shape=None, image_shape=None):
+    """
+    Get radial depth image :math:`\mathbf{X}_{rd}\in\mathbb{R}^{img_shape×1}` from depth scaled homogeneous pixel
+    co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{img_shape×3}`.\n
+
+    :param ds_pixel_coords: Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,1]*
+    :type ds_pixel_coords: array
+    :param inv_calib_mat: Inverse calibration matrix *[batch_shape,3,3]*
+    :type inv_calib_mat: array
+    :param batch_shape: Shape of batch. Inferred from inputs if None.
+    :type batch_shape: sequence of ints, optional
+    :param image_shape: Image shape. Inferred from inputs in None.
+    :type image_shape: sequence of ints, optional
+    :return: Radial depth image *[batch_shape,image_shape,1]*
+    """
+
+    if batch_shape is None:
+        batch_shape = inv_calib_mat.shape[:-2]
+    num_batch_dims = len(batch_shape)
+
+    if image_shape is None:
+        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]
+
+    # shapes as list
+    batch_shape = list(batch_shape)
+    image_shape = list(image_shape)
+
+    # BS x IS x 3
+    cam_coords = ds_pixel_to_cam_coords(ds_pixel_coords, inv_calib_mat, batch_shape, image_shape)[..., 0:3]
+
+    # BS x IS x 1
     return _ivy.reduce_sum(cam_coords**2, -1, keepdims=True)**0.5
 
 
@@ -318,44 +353,45 @@ def cam_to_ds_pixel_coords(coords_wrt_cam, calib_mat, batch_shape=None, image_di
     return _ivy_pg.transform(coords_wrt_cam, calib_mat, batch_shape, image_dims)
 
 
-def ds_pixel_to_cam_coords(ds_pixel_coords, inv_calib_mat, batch_shape=None, image_dims=None, dev_str=None):
+def ds_pixel_to_cam_coords(ds_pixel_coords, inv_calib_mat, batch_shape=None, image_shape=None, dev_str=None):
     """
-    Get camera-centric homogeneous co-ordinates image :math:`\mathbf{X}_c\in\mathbb{R}^{h×w×4}` from
-    depth scaled homogeneous pixel co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{h×w×3}`.\n
+    Get camera-centric homogeneous co-ordinates image :math:`\mathbf{X}_c\in\mathbb{R}^{img_shape×4}` from
+    depth scaled homogeneous pixel co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{img_shape×3}`.\n
     `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_
     page 155, matrix inverse of equation 6.3
 
-    :param ds_pixel_coords: Depth scaled homogeneous pixel co-ordinates image *[batch_shape,h,w,3]*
+    :param ds_pixel_coords: Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shap,3]*
     :type ds_pixel_coords: array
     :param inv_calib_mat: Inverse calibration matrix *[batch_shape,3,3]*
     :type inv_calib_mat: array
     :param batch_shape: Shape of batch. Inferred from inputs if None.
     :type batch_shape: sequence of ints, optional
-    :param image_dims: Image dimensions. Inferred from inputs in None.
-    :type image_dims: sequence of ints, optional
+    :param image_shape: Image dimensions. Inferred from inputs in None.
+    :type image_shape: sequence of ints, optional
     :param dev_str: device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc. Same as x if None.
     :type dev_str: str, optional
-    :return: Camera-centric homogeneous co-ordinates image *[batch_shape,h,w,4]*
+    :return: Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
     """
 
     if batch_shape is None:
-        batch_shape = ds_pixel_coords.shape[:-3]
+        batch_shape = inv_calib_mat.shape[:-2]
+    num_batch_dims = len(batch_shape)
 
-    if image_dims is None:
-        image_dims = ds_pixel_coords.shape[-3:-1]
+    if image_shape is None:
+        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]
 
     if dev_str is None:
         dev_str = _ivy.dev_str(ds_pixel_coords)
 
     # shapes as list
     batch_shape = list(batch_shape)
-    image_dims = list(image_dims)
+    image_shape = list(image_shape)
 
-    # BS x H x W x 3
-    cam_coords = _ivy_pg.transform(ds_pixel_coords, inv_calib_mat, batch_shape, image_dims)
+    # BS x IS x 3
+    cam_coords = _ivy_pg.transform(ds_pixel_coords, inv_calib_mat, batch_shape, image_shape)
 
-    # BS x H x W x 4
-    return _ivy.concatenate((cam_coords, _ivy.ones(batch_shape + image_dims + [1], dev_str=dev_str)), -1)
+    # BS x IS x 4
+    return _ivy_mec.make_coordinates_homogeneous(cam_coords, batch_shape + image_shape)
 
 
 def world_to_cam_coords(coords_wrt_world, ext_mat, batch_shape=None, image_dims=None, dev_str=None):
@@ -470,89 +506,94 @@ def world_to_ds_pixel_coords(coords_wrt_world, full_mat, batch_shape=None, image
     return _ivy_pg.transform(coords_wrt_world, full_mat, batch_shape, image_dims)
 
 
-def ds_pixel_to_world_coords(ds_pixel_coords, inv_full_mat, batch_shape=None, image_dims=None, dev_str=None):
+def ds_pixel_to_world_coords(ds_pixel_coords, inv_full_mat, batch_shape=None, image_shape=None):
     """
-    Get world-centric homogeneous co-ordinates image :math:`\mathbf{X}_w\in\mathbb{R}^{h×w×4}` from depth scaled
-    homogeneous pixel co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{h×w×3}`.\n
+    Get world-centric homogeneous co-ordinates image :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}` from depth scaled
+    homogeneous pixel co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`.\n
     `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_
     combination of page 155, matrix inverse of equation 6.3, and matrix inverse of page 156, equation 6.6
 
-    :param ds_pixel_coords: Depth scaled homogeneous pixel co-ordinates image: *[batch_shape,h,w,3]*
+    :param ds_pixel_coords: Depth scaled homogeneous pixel co-ordinates image: *[batch_shape,image_shape,3]*
     :type ds_pixel_coords: array
     :param inv_full_mat: Inverse full projection matrix *[batch_shape,3,4]*
     :type inv_full_mat: array
     :param batch_shape: Shape of batch. Inferred from inputs if None.
     :type batch_shape: sequence of ints, optional
-    :param image_dims: Image dimensions. Inferred from inputs in None.
-    :type image_dims: sequence of ints, optional
-    :param dev_str: device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc. Same as x if None.
-    :type dev_str: str, optional
-    :return: World-centric homogeneous co-ordinates image *[batch_shape,h,w,4]*
+    :param image_shape: Image shape. Inferred from inputs in None.
+    :type image_shape: sequence of ints, optional
+    :return: World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
     """
 
     if batch_shape is None:
-        batch_shape = ds_pixel_coords.shape[:-3]
+        batch_shape = inv_full_mat.shape[:-2]
+    num_batch_dims = len(batch_shape)
 
-    if image_dims is None:
-        image_dims = ds_pixel_coords.shape[-3:-1]
-
-    if dev_str is None:
-        dev_str = _ivy.dev_str(ds_pixel_coords)
+    if image_shape is None:
+        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]
 
     # shapes as list
     batch_shape = list(batch_shape)
-    image_dims = list(image_dims)
+    image_shape = list(image_shape)
 
-    # BS x H x W x 4
-    ds_pixel_coords = _ivy.concatenate((ds_pixel_coords, _ivy.ones(batch_shape + image_dims + [1], dev_str=dev_str)), -1)
+    # BS x IS x 4
+    ds_pixel_coords = _ivy_mec.make_coordinates_homogeneous(ds_pixel_coords, batch_shape)
 
-    # BS x H x W x 3
-    world_coords = _ivy_pg.transform(ds_pixel_coords, inv_full_mat, batch_shape, image_dims)
+    # BS x IS x 3
+    world_coords = _ivy_pg.transform(ds_pixel_coords, inv_full_mat, batch_shape, image_shape)
 
-    # BS x H x W x 4
-    return _ivy.concatenate((world_coords, _ivy.ones(batch_shape + image_dims + [1], dev_str=dev_str)), -1)
+    # BS x IS x 4
+    return _ivy_mec.make_coordinates_homogeneous(world_coords, batch_shape + image_shape)
 
 
-def ds_pixel_coords_to_world_ray_vectors(ds_pixel_coords, inv_full_mat, camera_center=None, batch_shape=None, image_dims=None):
+def pixel_coords_to_world_ray_vectors(inv_full_mat, pixel_coords=None, camera_center=None, batch_shape=None,
+                                      image_shape=None):
     """
-    Calculate world-centric ray vector image :math:`\mathbf{RV}\in\mathbb{R}^{h×w×3}` from homogeneous pixel co-ordinate
-    image :math:`\mathbf{X}_p\in\mathbb{R}^{h×w×3}`. Each ray vector :math:`\mathbf{rv}_{i,j}\in\mathbb{R}^{3}` is
+    Calculate world-centric ray vector image :math:`\mathbf{RV}\in\mathbb{R}^{is×3}` from homogeneous pixel co-ordinate
+    image :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`. Each ray vector :math:`\mathbf{rv}_{i,j}\in\mathbb{R}^{3}` is
     represented as a unit vector from the camera center :math:`\overset{\sim}{\mathbf{C}}\in\mathbb{R}^{3×1}`, in the
     world frame. Co-ordinates :math:`\mathbf{x}_{i,j}\in\mathbb{R}^{3}` along the world ray can then be parameterized as
     :math:`\mathbf{x}_{i,j}=\overset{\sim}{\mathbf{C}} + λ\mathbf{rv}_{i,j}`, where :math:`λ` is a scalar who's
     magnitude dictates the position of the world co-ordinate along the world ray.
 
-    :param ds_pixel_coords: Homogeneous pixel co-ordinates image *[batch_shape,h,w,3]*
-    :type ds_pixel_coords: array
     :param inv_full_mat: Inverse full projection matrix *[batch_shape,3,4]*
     :type inv_full_mat: array
+    :param pixel_coords: Homogeneous pixel co-ordinates image, created uniformly if None. *[batch_shape,image_shape,3]*
+    :type pixel_coords: array, optional
     :param camera_center: Camera centers, inferred from inv_full_mat if None *[batch_shape,3,1]*
     :type camera_center: array, optional
     :param batch_shape: Shape of batch. Inferred from inputs if None.
     :type batch_shape: sequence of ints, optional
-    :param image_dims: Image dimensions. Inferred from inputs in None.
-    :type image_dims: sequence of ints, optional
-    :return: World ray vectors *[batch_shape,h,w,3]*
+    :param image_shape: Image shape. Inferred from inputs in None.
+    :type image_shape: sequence of ints, optional
+    :return: World ray vectors *[batch_shape,image_shape,3]*
     """
 
     if batch_shape is None:
-        batch_shape = ds_pixel_coords.shape[:-3]
+        batch_shape = inv_full_mat.shape[:-2]
+    num_batch_dims = len(batch_shape)
 
-    if image_dims is None:
-        image_dims = ds_pixel_coords.shape[-3:-1]
+    if image_shape is None:
+        if pixel_coords is None:
+            raise Exception('if pixel_coords is not specified, image_shape must be specified when calling'
+                            'pixel_coords_to_world_ray_vectors')
+        image_shape = pixel_coords.shape[num_batch_dims:-1]
+    num_image_dims = len(image_shape)
 
     # shapes as list
     batch_shape = list(batch_shape)
-    image_dims = list(image_dims)
+    image_shape = list(image_shape)
 
     if camera_center is None:
         camera_center = inv_ext_mat_to_camera_center(inv_full_mat)
 
-    # BS x 1 x 1 x 3
-    camera_centers_reshaped = _ivy.reshape(camera_center, batch_shape + [1, 1, 3])
+    if pixel_coords is None:
+        pixel_coords = create_uniform_pixel_coords_image(image_shape, batch_shape, dev_str=_ivy.dev_str(inv_full_mat))
 
-    # BS x H x W x 3
-    vectors = ds_pixel_to_world_coords(ds_pixel_coords, inv_full_mat, batch_shape, image_dims)[..., 0:3] \
+    # BS x [1]xNID x 3
+    camera_centers_reshaped = _ivy.reshape(camera_center, batch_shape + [1]*num_image_dims + [3])
+
+    # BS x IS x 3
+    vectors = ds_pixel_to_world_coords(pixel_coords, inv_full_mat, batch_shape, image_shape)[..., 0:3] \
               - camera_centers_reshaped
 
     # BS x H x W x 3
@@ -797,7 +838,7 @@ def sphere_to_cam_coords(sphere_coords, forward_facing_z=True, batch_shape=None,
             (cam_coords[..., 1:2], cam_coords[..., 2:3], cam_coords[..., 0:1]), -1)
 
     # BS x H x W x 4
-    return _ivy.concatenate((cam_coords, _ivy.ones(batch_shape + image_dims + [1], dev_str=dev_str)), -1)
+    return _ivy_mec.make_coordinates_homogeneous(cam_coords, batch_shape + image_dims)
 
 
 def sphere_to_ds_pixel_coords(sphere_coords, calib_mat, batch_shape=None, image_dims=None):
@@ -942,6 +983,43 @@ def focal_lengths_and_pp_offsets_to_intrinsics_object(focal_lengths, pp_offsets,
     return intrinsics
 
 
+def calib_mat_to_intrinsics_object(calib_mat, image_dims, batch_shape=None):
+    """
+    Create camera intrinsics object from calibration matrix.
+
+    :param calib_mat: Calibration matrices *[batch_shape,3,3]*
+    :type calib_mat: array
+    :param image_dims: Image dimensions. Inferred from inputs in None.
+    :type image_dims: sequence of ints
+    :param batch_shape: Shape of batch. Inferred from inputs if None.
+    :type batch_shape: sequence of ints, optional
+    :return: Camera intrinsics object
+    """
+
+    if batch_shape is None:
+        batch_shape = calib_mat.shape[:-2]
+
+    # shapes as list
+    batch_shape = list(batch_shape)
+    image_dims = list(image_dims)
+
+    # BS x 2
+    focal_lengths = _ivy.concatenate((calib_mat[..., 0, 0:1], calib_mat[..., 1, 1:2]), -1)
+
+    # BS x 2
+    persp_angles = focal_lengths_to_persp_angles(focal_lengths, image_dims)
+
+    # BS x 2
+    pp_offsets = _ivy.concatenate((calib_mat[..., 0, -1:], calib_mat[..., 1, -1:]), -1)
+
+    # BS x 3 x 3
+    inv_calib_mat = _ivy.inv(calib_mat)
+
+    # intrinsics object
+    intrinsics = _Intrinsics(focal_lengths, persp_angles, pp_offsets, calib_mat, inv_calib_mat)
+    return intrinsics
+
+
 def ext_mat_and_intrinsics_to_cam_geometry_object(ext_mat, intrinsics, batch_shape=None, dev_str=None):
     """
     Create camera geometry object from extrinsic matrix :math:`\mathbf{E}\in\mathbb{R}^{3×4}`, and camera intrinsics
